import os
import shutil
import tkinter


class Component:
    def __init__(self,
                 name: str,
                 path: str,
                 ):
        """
        Class that Folder inherits from, created in case there are other components to be tracked in folders in the
        future, such as files.

        :param str, name: name of the component
        :param str, path: path to save the component in. The last part of the path must be the name of the component
        """
        self.name = name
        self.path = path
        self.parent = None  # parent Component this component belongs to; can be set to None if it is the first
        # component being made

        try:
            self.save_to_disk()
        except FileExistsError as error:
            root = tkinter.Tk()
            yes_or_no_result = tkinter.messagebox.askyesno(
                f'Save to disk',
                f'File or folder at {path}. '
                f'Would you like to create one with "_copy_#" appended to the end of the file/folder name?')
            if yes_or_no_result is True:
                # rename the file something else
                for i in range(20):  # randomly put 20 here, just assuming there won't already be more than this
                    # number of folders that will be made with the same name
                    try:
                        old_name = self.get_name()
                        old_path = self.get_path()
                        new_name = old_name + f'_copy_{i}'
                        new_path = old_path + f'_copy_{i}'
                        self.set_name(name=new_name)
                        self.set_path(path=new_path)
                        self.save_to_disk()
                        root.destroy()
                        break
                    except FileExistsError as error:
                        self.set_name(name=old_name)
                        self.set_path(path=old_path)
            else:
                raise error

    def get_name(self):
        return self.name

    def get_path(self):
        return self.path

    def get_parent(self):
        return self.parent

    def set_name(self,
                 name: str,
                 ):
        self.name = name

    def set_path(self,
                 path: str,
                 ):
        self.path = path

    def set_parent(self,
                   component,
                   ):
        """
        Set the parent component for this component

        :param Component, component:
        :return:
        """
        self.parent = component

    def save_to_disk(self):
        """
        Actually create the component on the computer at its path with its name
        :return:
        """

        raise NotImplementedError

    def delete_from_disk(self):
        """
        Actually delete the component from the computer
        :return:
        """
        shutil.rmtree(path=self.path)


class Folder(Component):
    """
    Class to create and store a folder hierarchy and to easily create folders and access paths to save files and
    folders in

    Example using the Folder class - run each line one at a time to see changes to disk:

    root_path = os.getcwd()

    test_folder_name = 'test'
    test_folder_path = os.path.join(root_path, test_folder_name)

    test_folder = Folder(folder_name=test_folder_name, folder_path=test_folder_path)

    sub_folder_one = test_folder.make_and_add_sub_folder(sub_folder_name='sub_folder_one')
    sub_folder_two = test_folder.make_and_add_sub_folder(sub_folder_name='sub_folder_two')
    sub_sub_folder_one = sub_folder_one.make_and_add_sub_folder(sub_folder_name='sub_sub_folder_one')

    sub_folder_one.delete_from_disk()
    sub_folder_two.delete_from_disk()

    test_folder_two = Folder(folder_name=test_folder_name, folder_path=test_folder_path)

    test_folder.delete_from_disk()
    test_folder_two.delete_from_disk()

    """

    def __init__(self,
                 folder_name: str,
                 folder_path: str,
                 ):
        """
        A folder can have files and folders in it, but for now just care about it containing folders. The name of the
        folder should be the same as the last part of the folder path.

        :param str, folder_path: path to save the folder on disk
        :param str, folder_name: Should be the same as the last part of the folder path
        """
        super().__init__(
            name=folder_name,
            path=folder_path
        )
        self.children = set()

    def get_name(self):
        return super().get_name()

    def get_path(self):
        return super().get_path()

    def get_parent(self):
        return super().get_parent()

    def save_to_disk(self):
        os.makedirs(
            name=self.path,
        )

    def delete_from_disk(self):
        super().delete_from_disk()

    def set_parent(self,
                   component: Component,
                   ):
        super().set_parent(component=component)

    def get_parent(self):
        super().get_parent()

    def get_children(self):
        return self.children

    def add_child_component(self,
                            component: Component,
                            ):
        """
        Add a component to the set of children and set the parent of the child component to be this folder

        :param Component, component:
        :return:
        """
        self.children.add(component)
        component.set_parent(component=self)

    def remove_and_delete_component(self,
                                    component: Component,
                                    ):
        """
        Remove a child component from the children set and delete it from disk

        :param Component, component:
        :return:
        """
        self.children.remove(component)
        component.delete_from_disk()

    def make_and_add_sub_folder(self,
                                sub_folder_name: str,
                                ):
        """
        Create a sub-folder with a given name under the main folder; the path of the sub-folder is the name of the
        sub-folder concatenated on to the end of the path of the main folder

        :return: Folder, sub_folder: the sub-folder that was created
        """

        if sub_folder_name in self.children:
            raise Exception(f'Main folder already has a sub-folder called {sub_folder_name}')
        else:
            parent_folder_path = self.path
            sub_folder_path = os.path.join(
                parent_folder_path,
                sub_folder_name
            )
            sub_folder = Folder(
                folder_name=sub_folder_name,
                folder_path=sub_folder_path,
            )
            sub_folder.set_parent(component=self)

            self.add_child_component(component=sub_folder)

        return sub_folder



from typing import Union, List
from logging import StreamHandler, LogRecord, Formatter
from .bots import SlackBot, RTMSlackBot


# some predefined formatters for convenience
# detailed message with time, name, level, and message
DETAILED_FORMATTER = Formatter(
    '`%(asctime)s` `%(name)s` *%(levelname)s*: %(message)s'
)

# timestamped message
TIMESTAMPED_FORMATTER = Formatter(
    '`%(asctime)s` %(message)s'
)


class SlackLoggingHandler(StreamHandler):
    # default formatter
    DEFAULT_FORMATTER = TIMESTAMPED_FORMATTER

    def __init__(self,
                 token: str,
                 channel_name: str,
                 user_member_ids: Union[str, List[str]] = None,
                 alert_level: int = 30,
                 formatter: Formatter = None,
                 **kwargs,
                 ):
        """
        A logging stream handler which logs to a slack channel. This class encompasses the RTMSlackBot class to enable
        real-time-messaging handling using the connected slack bot.

        :param str token: token to connect to slack client
        :param str channel_name: channel to message on. for example, #channelname
        :param user_member_ids: user member ID(s) to tag if something goes wrong (if logging level exceeds alert level)
        :param alert_level: alert level to notify the defined user at
        :param formatter: logging formatter to use (if not specified, the class DEFAULT_FORMATTER is used)
        """
        StreamHandler.__init__(self)
        if formatter is None:
            formatter = self.DEFAULT_FORMATTER
        self.setFormatter(formatter)
        self.alert_level = alert_level
        self.bot = SlackBot(
            user_member_ids=user_member_ids,
            token=token,
            channel_name=channel_name,
            **kwargs,
        )

    def emit(self, record: LogRecord) -> None:
        # catch and format message
        msg = self.format(record)
        # determine whether to tag admin
        tag_admin = True if record.levelno >= self.alert_level else False
        self.bot.post_slack_message(
            msg,
            tag_admin=tag_admin,
        )


class RTMSlackLoggingHandler(StreamHandler):
    # default formater
    DEFAULT_FORMATTER = TIMESTAMPED_FORMATTER

    def __init__(self,
                 token: str,
                 channel_name: str,
                 user_member_ids: Union[str, List[str]] = None,
                 alert_level: int = 30,
                 formatter: Formatter = None,
                 **kwargs,
                 ):
        """
        A logging stream handler which logs to a slack channel. This class encompasses the RTMSlackBot class to enable
        real-time-messaging handling using the connected slack bot.

        :param str token: token to connect to slack client
        :param bot_name: name for the bot
        :param str channel_name: channel to message on. for example, #channelname
        :param user_member_ids: user member ID(s) to tag if something goes wrong (if logging level exceeds alert level)
        :param alert_level: alert level to notify the defined user at
        """
        StreamHandler.__init__(self)
        if formatter is None:
            formatter = self.DEFAULT_FORMATTER
        self.setFormatter(formatter)
        self.alert_level = alert_level
        self.bot = RTMSlackBot(
            user_member_ids=user_member_ids,
            token=token,
            channel_name=channel_name,
            **kwargs,
        )

    def emit(self, record: LogRecord) -> None:
        # catch and format message
        msg = self.format(record)
        # determine whether to tag admin
        tag_admin = True if record.levelno >= self.alert_level else False
        self.bot.post_slack_message(
            msg,
            tag_admin=tag_admin,
        )

    def run_on(self, *, event: str):
        """a pass-through for the RTMClient.run_on decorator"""
        return self.bot.rtm_client.run_on(event=event)

    def start_rtm_thread(self):
        """starts the RTM monitor thread"""
        self.bot.start_rtm_client()



from datetime import datetime


_unit_seconds = 'seconds'
_unit_minutes = 'minutes'
_unit_hours = 'hours'

_time_units = {0: _unit_seconds,
               1: _unit_minutes,
               2: _unit_hours,
               }

_reverse_time_units = dict((v, k) for k, v in _time_units.items())

_default_datetime_format = '%Y_%m_%d_%H_%M_%S'  # to add in milliseconds, add %f


def array_of_str_to_array_of_datetime_objects(list_of_string_values,
                                              datetime_format: str =_default_datetime_format,
                                              ):
    """
    Convert a list of string values, back into datetime objects with a specific format; in this case, the string has
    to have been a datetime object that was converted into a string with the datetime_format that is passed in this
    function.

    Main use has previously been when files where timestamped, and when the file names need to be converted back into
    datetime objects in order to do calculations

    :param list_of_string_values:
    :param str, datetime_format: a string to represent the datetime format the string should be converted into; it
        should also have been the format that the strings already are in
    :return:
    """
    array_of_datetime_objects = []
    for value in list_of_string_values:
        datetime_object = datetime.strptime(value, datetime_format)
        array_of_datetime_objects.append(datetime_object)
    return array_of_datetime_objects


def array_of_datetime_objects_to_array_of_relative_datetime_objects(array_of_datetime_objects,
                                                                    units,
                                                                    ):
    """
    Convert an array of datetime objects that are absolute times, and return an array where all the times in the
    array are relative to the first time in the array. The relativity can be in secons, minutes, or hours.

    :param array_of_datetime_objects:
    :param units: one of _unit_seconds, _unit_minutes, or _unit_hours, which are defined in this datetime_utils module
    :return:
    """
    # takes a list of datetime objects, and makes all the values relative to the first object in the list as the
    # 0 points. units must be one of the _time_units types (_unit_seconds, _unit_minutes, or _unit_hours)

    # make an array of timedelta objects where each value is the difference between the actual time relative to
    # the first time point
    array_of_datetime_timedelta = [datetime_value - array_of_datetime_objects[0] for datetime_value in
                                   array_of_datetime_objects]

    # convert the relative timedeltas to floats, where the float number is the number of seconds since the first
    # time point
    array_of_relative_x_in_seconds = [array_of_datetime_timedelta[index].total_seconds() for index
                                      in range(len(array_of_datetime_timedelta))]

    if units == _unit_seconds:
        array_of_relative_datetime_objects = array_of_relative_x_in_seconds
    elif units == _unit_minutes:
        array_of_relative_x_in_minutes = [array_of_relative_x_in_seconds[index] / 60.0 for index in
                                          range(len(array_of_relative_x_in_seconds))]
        array_of_relative_datetime_objects = array_of_relative_x_in_minutes
    elif units == _unit_hours:
        array_of_relative_x_in_hours = [array_of_relative_x_in_seconds[index] / 3600.0 for index in
                                        range(len(array_of_relative_x_in_seconds))]
        array_of_relative_datetime_objects = array_of_relative_x_in_hours

    return array_of_relative_datetime_objects

